/*
 *  Copyright 2020-2023 Google LLC
 *  Copyright 2020-2023 EPAM Systems, Inc
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.opengroup.osdu.notification.pubsub.receiver;

import java.util.Map;
import lombok.extern.slf4j.Slf4j;
import org.opengroup.osdu.oqm.core.model.OqmAckReplier;
import org.opengroup.osdu.oqm.core.model.OqmMessage;
import org.opengroup.osdu.oqm.core.model.OqmMessageReceiver;
import org.opengroup.osdu.notification.service.ExternalSubscriptionsManager;
import org.opengroup.osdu.notification.thread.ThreadScopeContextHolder;

@Slf4j
public class OqmConfigurationEventReceiver implements OqmMessageReceiver {

    private static final String SUBSCRIPTION_CREATED = "Subscription Created";
    private static final String SUBSCRIPTION_UPDATED = "Subscription Updated";
    private static final String SUBSCRIPTION_DELETED = "Subscription Deleted";

    private final ExternalSubscriptionsManager externalSubscriptionsManager;

    public OqmConfigurationEventReceiver(ExternalSubscriptionsManager externalSubscriptionsManager) {
        this.externalSubscriptionsManager = externalSubscriptionsManager;
    }

    @Override
    public void receiveMessage(OqmMessage message, OqmAckReplier replier) {
        try {
            handleMessage(message);
        } catch (Exception e) {
            log.error("OQM | Control topic | Message not acknowledged by client", e);
            replier.nack();
        } finally {
            ThreadScopeContextHolder.currentThreadScopeAttributes().clear();
        }
        log.debug("OQM | Control topic | Message acknowledged by client");
        replier.ack();
    }

    private void handleMessage(OqmMessage oqmMessage) {
        String pubsubMessage = oqmMessage.getData();
        Map<String, String> headerAttributes = oqmMessage.getAttributes();
        String subscriptionId = headerAttributes.get("subscription-id");
        String serviceTopic = headerAttributes.get("topic");
        String dataPartitionId = headerAttributes.get("data-partition-id");
        log.debug("OQM | Control topic | Received message: `{}` for service topic: `{}` with Sub ID: `{}`", pubsubMessage, serviceTopic, subscriptionId);

        switch (pubsubMessage) {
            case SUBSCRIPTION_CREATED, SUBSCRIPTION_UPDATED -> {
                externalSubscriptionsManager.updateSubscription(dataPartitionId, subscriptionId);
                log.debug(
                    "OQM | Control topic | Subscription info with Sub ID: `{}` and topic: `{}` was added to cache.",
                    subscriptionId, serviceTopic);
            }
            case SUBSCRIPTION_DELETED -> {
                boolean deleted = externalSubscriptionsManager.removeSubscription(dataPartitionId,
                    serviceTopic, subscriptionId);
                if (deleted) {
                    log.debug(
                        "OQM | Control topic | Subscription info with Sub ID: `{}` and topic: `{}` was removed from cache.",
                        subscriptionId, serviceTopic);
                } else {
                    log.warn(
                        "OQM | Control topic requested to delete subscription with id: `{}`, but it was not found in cache.",
                        subscriptionId);
                }
            }
            default -> log.warn(
                "Not supported operation! OQM | Control topic | Received unsupported control topic operation: {}",
                pubsubMessage);
        }
    }
}