/*
 *   Copyright 2017-2020, Schlumberger
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package org.opengroup.osdu.notification.auth;

import org.apache.commons.lang3.StringUtils;
import org.opengroup.osdu.core.common.cache.ICache;
import org.opengroup.osdu.core.common.model.entitlements.AuthorizationResponse;
import org.opengroup.osdu.core.common.model.entitlements.Groups;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.provider.interfaces.IAuthorizationService;
import org.opengroup.osdu.core.common.util.Crc32c;
import org.opengroup.osdu.notification.di.RequestInfoExt;
import org.opengroup.osdu.notification.utils.Config;
import org.opengroup.osdu.notification.provider.interfaces.IServiceAccountValidator;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.web.context.annotation.RequestScope;

import jakarta.servlet.http.HttpServletRequest;
import java.util.Arrays;

@Component("authorizationFilter")
@RequestScope
public class AuthorizationFilter {
    private static final String BEARER_PREFIX = "Bearer ";

    @Autowired
    private HttpServletRequest request;
    @Autowired
    private IAuthorizationService authService;
    @Autowired
    private RequestInfoExt requestInfoExt;
    @Autowired
    private IServiceAccountValidator validator;
    @Autowired
    private ICache<String, Groups> cache;

    public boolean hasAnyPermission(String... requiredRoles) {
        DpsHeaders dpsHeaders = requestInfoExt.getHeaders();
        if (StringUtils.isBlank(dpsHeaders.getAuthorization()))
            throw new AppException(302, "Redirect", "No credentials sent on request.");

        if (StringUtils.isBlank(dpsHeaders.getPartitionId())) {
            requestInfoExt.assignPartitionIdIfNotInHeader();
        }

        String path = request.getServletPath();
        if ("GET".equals(request.getMethod()) &&
           (path.equals("/swagger-ui.html") || path.equals("/info"))) {
            return true;
        }

        if (Arrays.asList(requiredRoles).contains(Config.CRON) && requestInfoExt.isCronRequest()) {
            dpsHeaders.put(DpsHeaders.USER_EMAIL, Config.CRON);
            requestInfoExt.setHeaders(dpsHeaders);
        } else if (Arrays.asList(requiredRoles).contains(Config.PUBSUB)) {
            String jwt = dpsHeaders.getAuthorization().substring(BEARER_PREFIX.length());
            if (!this.validator.isValidPublisherServiceAccount(jwt)) {
                return this.authorizeWithCacheOrEntitlements(requiredRoles, dpsHeaders);
            } else {
                return false;
            }
        } else {
            return this.authorizeWithCacheOrEntitlements(requiredRoles, dpsHeaders);
        }
        return true;
    }

    private boolean authorizeWithCacheOrEntitlements(String[] requiredRoles, DpsHeaders dpsHeaders) {
        String cacheKey = getGroupCacheKey(dpsHeaders);
        Groups groups = cache.get(cacheKey);
        if(groups == null) {
            AuthorizationResponse authorizationResponse = this.authorizeWithEntitlements(requiredRoles, dpsHeaders);
            cache.put(cacheKey, authorizationResponse.getGroups());
            return true;
        } else {
            return authorizeWithCache(requiredRoles, groups);
        }
    }

    private boolean authorizeWithCache(String[] requiredRoles, Groups groups) {
        if(groups.any(requiredRoles)) {
            return true;
        }
        return false;
    }

    private String getGroupCacheKey(DpsHeaders dpsHeaders) {
        String key = String.format("notification-entitlement-groups:%s:%s", dpsHeaders.getPartitionIdWithFallbackToAccountId(),
                dpsHeaders.getAuthorization());
        return Crc32c.hashToBase64EncodedString(key);
    }

    private AuthorizationResponse authorizeWithEntitlements(String[] requiredRoles, DpsHeaders dpsHeaders) {
        AuthorizationResponse authorizationResponse = authService.authorizeAny(dpsHeaders, requiredRoles);
        dpsHeaders.put(DpsHeaders.USER_EMAIL, authorizationResponse.getUser());
        requestInfoExt.setHeaders(dpsHeaders);
        return authorizationResponse;
    }
}
