/* Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.notification.subscriptions;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import com.amazonaws.services.sqs.model.Message;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.gson.JsonArray;
import com.google.gson.JsonObject;
import com.sun.jersey.api.client.Client;
import com.sun.jersey.api.client.ClientResponse;
import com.sun.jersey.api.client.WebResource;
import org.apache.commons.lang3.tuple.Pair;
import org.opengroup.osdu.core.common.model.entitlements.Acl;
import org.opengroup.osdu.core.common.model.file.LocationResponse;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.legal.Legal;
import org.opengroup.osdu.notification.util.AwsTestUtils;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import javax.ws.rs.core.MediaType;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.SecureRandom;
import java.security.cert.X509Certificate;
import java.time.Instant;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

public class FileTestUtils {

    private final DpsHeaders dpsHeaders;
    private final Client restClient;
    private final String baseUrl;
    private final static String DATASET_KIND = "osdu:wks:dataset--File.Generic:1.0.0";
    private final ObjectMapper objectMapper = new ObjectMapper();

    private static class NullTrustManager implements X509TrustManager {
        @Override
        public X509Certificate[] getAcceptedIssuers() {
            return null;
        }

        @Override
        public void checkClientTrusted(X509Certificate[] certs, String authType) {
        }

        @Override
        public void checkServerTrusted(X509Certificate[] certs, String authType) {
        }
    }

    private record FileMetadataResponse(String id) {
    }

    public FileTestUtils(DpsHeaders headers) {
        TrustManager[] trustAllCerts = new TrustManager[]{new NullTrustManager()};

        try {
            SSLContext sc = SSLContext.getInstance("TLS");
            sc.init(null, trustAllCerts, new SecureRandom());
            HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());
        } catch (Exception e) {
        }
        restClient = Client.create();
        dpsHeaders = headers;
        baseUrl = System.getProperty("FILE_URL", System.getenv("FILE_URL"));
        assertNotEquals(null, baseUrl, "Must set the FILE_URL environment variable!");
    }

    /**
     * Returns the expected message values for the given the legal tag name to use.
     * @param legalTagName The legal tag name.
     * @return The expected message values. Used to compare with the actual message values.
     */
    public String uploadFile(String legalTagName) throws JsonProcessingException {
        LocationResponse locationResponse = getFileLocation();
        String signedUrl = locationResponse.getLocation().get("SignedURL");
        assertNotEquals(null, signedUrl);
        String fileSource = locationResponse.getLocation().get("FileSource");
        uploadTestFile(signedUrl);
        return setFileMetadata(fileSource, legalTagName);
    }

    private String getFileMetadataURL() {
        try {
            URL mergedURL = new URL(baseUrl + "/files/metadata");
            return mergedURL.toString();
        } catch (MalformedURLException e) {
            fail("Failed to get file metadata URL");
        }
        return null;
    }

    private String getFileUploadURL() {
        try {
            URL mergedURL = new URL(baseUrl + "/files/uploadURL");
            return mergedURL.toString();
        } catch (MalformedURLException e) {
            fail("Failed to get file upload URL");
        }
        return null;
    }

    private ClientResponse sendRequest(String url, String method, boolean addHeaders, Object body, MediaType contentType) {
        WebResource resource = restClient.resource(url);
        WebResource.Builder resourceBuilder = resource.getRequestBuilder();
        if (contentType != null) {
            resourceBuilder.accept(contentType);
            resourceBuilder.type(contentType);
        }
        if (addHeaders) {
            dpsHeaders.getHeaders().forEach(resourceBuilder::header);
        }
        return (body == null) ? resourceBuilder.method(method, ClientResponse.class) :  resourceBuilder.method(method, ClientResponse.class, body);
    }

    private LocationResponse getFileLocation() throws JsonProcessingException {
        ClientResponse response = sendRequest(getFileUploadURL(), "GET", true, null, null);
        assertEquals(200, response.getStatus());
        String responseData = response.getEntity(String.class);
        return objectMapper.readValue(responseData, LocationResponse.class);
    }

    private void uploadTestFile(String url) {
        ClientResponse response = sendRequest(url, "PUT", false, "Test file upload.", MediaType.MULTIPART_FORM_DATA_TYPE);
        assertEquals(200, response.getStatus());
    }

    private JsonObject getFileMetadataData(String fileSource) {
        JsonObject dataObject = new JsonObject();
        dataObject.addProperty("Endian", "BIG");
        dataObject.addProperty("Description", "Test file");
        dataObject.addProperty("TotalSize", "1048576");
        JsonObject fileSourceInfo = new JsonObject();
        fileSourceInfo.addProperty("FileSource", fileSource);
        fileSourceInfo.addProperty("Name", "1234");
        fileSourceInfo.addProperty("PreloadFilePath", "gs://osdu-cicd-epam-persistent-area/c3af38c1-654d-47a0-a3e6-9e94c32add84/b62b104f843142f49ee6d747e6bdd49d");
        fileSourceInfo.addProperty("PreloadFileCreateUser", "user1");
        fileSourceInfo.addProperty("PreloadFileModifyDate", "mar 11");
        fileSourceInfo.addProperty("PreloadFileModifyUser", "mar 11");
        JsonObject datasetProperties = new JsonObject();
        datasetProperties.add("FileSourceInfo", fileSourceInfo);
        dataObject.add("DatasetProperties", datasetProperties);
        return dataObject;
    }

    private JsonArray getJsonArrayFromStringArray(Iterable<String> input) {
    	JsonArray jsonArray = new JsonArray();
        for (String str : input) {
            jsonArray.add(str);
        }
        return jsonArray;
    }

    private JsonObject getFileMetadataBase(String legalTagName) {
        Pair<Acl, Legal> aclLegalPair = RecordUtils.getAclAndLegal(RecordUtils.getDefaultAcls(true), legalTagName, "US");
        JsonObject jsonAcl = new JsonObject();
        Acl acl = aclLegalPair.getLeft();
        jsonAcl.add("viewers", getJsonArrayFromStringArray(Arrays.asList(acl.viewers)));
        jsonAcl.add("owners", getJsonArrayFromStringArray(Arrays.asList(acl.owners)));

        Legal legal = aclLegalPair.getRight();
        JsonObject jsonLegal = new JsonObject();
        jsonLegal.add("legaltags", getJsonArrayFromStringArray(legal.getLegaltags()));
        jsonLegal.add("otherRelevantDataCountries", getJsonArrayFromStringArray(legal.getOtherRelevantDataCountries()));
        jsonLegal.addProperty("status", "compliant");

        JsonObject fileMetadata = new JsonObject();
        fileMetadata.add("acl", jsonAcl);
        fileMetadata.add("legal", jsonLegal);
        return fileMetadata;
    }

    private JsonObject getFileMetadata(String fileSource, String legalTagName) {
        JsonObject fileMetadata = getFileMetadataBase(legalTagName);
        fileMetadata.addProperty("createUser", "osdu-community-sa-airflow@nice-etching-277309.iam.gserviceaccount.com");
        fileMetadata.addProperty("createTime", "2021-02-22T18:50:47.498Z");
        fileMetadata.addProperty("modifyUser", "osdu-community-sa-airflow@nice-etching-277309.iam.gserviceaccount.com");
        fileMetadata.addProperty("modifyTime", "2021-02-22T21:13:10.587Z");
        fileMetadata.addProperty("kind", DATASET_KIND);

        JsonObject dataObject = getFileMetadataData(fileSource);
        fileMetadata.add("data", dataObject);

        return fileMetadata;
    }

    private String setFileMetadata(String fileSource, String legalTagName) throws JsonProcessingException {
        JsonObject fileMetadata = getFileMetadata(fileSource, legalTagName);

        ClientResponse response = sendRequest(getFileMetadataURL(), "POST", true, fileMetadata.toString(), null);
        assertEquals(201, response.getStatus());
        String responseData = response.getEntity(String.class);
        return objectMapper.readValue(responseData, FileMetadataResponse.class).id();
    }

    private void compareVersionIdWithExpected(String fileId, Object versionId, Instant started, Instant ended) {
        assertNotEquals(null, versionId);
        assertTrue(versionId instanceof String);
        String versionString = versionId.toString();
        assertTrue(versionString.startsWith(fileId));
        String versionIdString = versionString.substring(fileId.length() + 1);
        long versionIdMicros = Long.parseLong(versionIdString);
        final long MICROS_PER_SECOND = 1000000;
        final long NANOS_PER_MICRO = 1000;
        long versionIdNanos = (versionIdMicros % MICROS_PER_SECOND) * NANOS_PER_MICRO;
        long versionIdSeconds = versionIdMicros / MICROS_PER_SECOND;
        Instant versionIdInstant = Instant.ofEpochSecond(versionIdSeconds, versionIdNanos);
        assertTrue(versionIdInstant.isAfter(started));
        assertTrue(versionIdInstant.isBefore(ended));
    }

    private void compareTimestamp(Object timestampObj, Instant started, Instant ended) {
        assertNotEquals(null, timestampObj);
        assertTrue(timestampObj instanceof Number);
        Instant instantTs = Instant.ofEpochMilli(((Number) timestampObj).longValue());
        assertTrue(instantTs.isAfter(started));
        assertTrue(instantTs.isBefore(ended));
    }

    private void checkDatasetDetails(Map<String, Object> properties, String fileId, Instant started, Instant ended) {
        assertEquals(fileId, properties.get("datasetId"));
        compareVersionIdWithExpected(fileId, properties.get("datasetVersionId"), started, ended);
        assertEquals("FILE", properties.get("datasetType").toString());
        Object recordCount = properties.get("recordCount");
        assertTrue(recordCount instanceof Number);
        assertEquals(1, recordCount);
        compareTimestamp(properties.get("timestamp"), started, ended);
    }

    private void checkInProgressStatus(Map<String, Object> properties, Instant started, Instant ended) {
        Object errorCode = properties.get("errorCode");
        assertNotEquals(null, errorCode);
        assertTrue(errorCode instanceof Number);
        assertEquals(0, errorCode);
        assertEquals(dpsHeaders.getUserId(), properties.get("userEmail"));
        compareTimestamp(properties.get("timestamp"), started, ended);
        Object message = properties.get("message");
        assertNotEquals(null, message);
        assertTrue(message instanceof String);
    }

    private void checkSuccessStatus(Map<String, Object> properties, String fileId, Instant started, Instant ended) {
        checkInProgressStatus(properties, started, ended);
        assertEquals(fileId, properties.get("recordId"));
        compareVersionIdWithExpected(fileId, properties.get("recordIdVersion"), started, ended);
    }

    public void compareMessagesWithExpected(String fileId, List<Message> receivedMessages, Instant started) throws Exception {
        Instant ended = Instant.now();
        assertEquals("Expect exactly 3 messages: dataset sync start, dataset sync end, and dataset details", 3, receivedMessages.size());
        String correlationId = null;
        boolean syncStartFound = false;
        boolean syncEndFound = false;
        boolean datasetDetailsFound = false;
        for (Message message : receivedMessages) {
            FileMessage fileMessage = AwsTestUtils.unwrapFirstMessage(message, FileMessage[].class);
            Map<String, Object> properties = fileMessage.properties();
            if (correlationId == null) {
                correlationId = properties.get("correlationId").toString();
            } else {
                assertEquals(correlationId, properties.get("correlationId").toString());
            }


            if (fileMessage.kind().equals("datasetDetails")) {
                assertFalse(datasetDetailsFound);
                datasetDetailsFound = true;
                checkDatasetDetails(properties, fileId, started, ended);
            } else if (fileMessage.kind().equals("status")) {
                Object status = properties.get("status");
                Object stage = properties.get("stage");
                assertNotEquals(null, status);
                assertNotEquals(null, stage);
                assertTrue(status instanceof String);
                assertTrue(stage instanceof String);
                assertEquals("DATASET_SYNC", stage);
                if (status.equals("IN_PROGRESS")) {
                    assertFalse(syncStartFound);
                    syncStartFound = true;
                    checkInProgressStatus(properties, started, ended);
                } else if (status.equals("SUCCESS")) {
                    assertFalse(syncEndFound);
                    syncEndFound = true;
                    checkSuccessStatus(properties, fileId, started, ended);
                } else {
                    fail("Unexpected status: " + status);
                }
            } else {
                fail("Unexpected kind: " + fileMessage.kind());
            }
        }
    }
}

record FileMessage (String kind, Map<String, Object> properties) {
}
