/* Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.notification.subscriptions;

import static org.junit.Assert.assertEquals;

import com.amazonaws.services.sqs.model.Message;
import com.google.gson.Gson;
import com.sun.jersey.api.client.Client;
import com.sun.jersey.api.client.ClientResponse;
import com.sun.jersey.api.client.WebResource;
import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.NoArgsConstructor;
import org.apache.commons.lang3.tuple.Pair;
import org.opengroup.osdu.core.common.http.DpsHeaderFactory;
import org.opengroup.osdu.core.common.logging.DefaultLogger;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.notification.util.AwsTestUtils;
import org.opengroup.osdu.notification.util.TestUtils;

import javax.ws.rs.core.MediaType;
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.util.List;
import java.util.stream.Collectors;

public class SchemaTestUtils {
    @Data
    @AllArgsConstructor
    @NoArgsConstructor
    public static class SchemaPubSubMessage {
        private String kind;
        private String op;
    }
    private final Client restClient;
    private final DpsHeaders dpsHeaders;
    private final String schemaApiBaseUrl;
    private final JaxRsDpsLog logger;
    public SchemaTestUtils(DpsHeaders headers) {
        restClient = Client.create();
        dpsHeaders = DpsHeaderFactory.createFromMap(headers.getHeaders());
        dpsHeaders.addCorrelationIdIfMissing();
        logger = new JaxRsDpsLog(new DefaultLogger(), dpsHeaders);
        schemaApiBaseUrl = String.format("%s/api/schema-service/v1", System.getProperty("SCHEMA_URL", System.getenv("SCHEMA_URL")));
    }

    private static final String SCHEMA_ID = "%s:%s:%s:%d.%d.%d";
    private static final String SCHEMA_BODY = """
{
    "schemaInfo":{
      "schemaIdentity":{
         "authority":"%s",
         "source":"%s",
         "entityType":"%s",
         "schemaVersionMajor":%d,
         "schemaVersionMinor":%d,
         "schemaVersionPatch":%d,
         "id":"%s"
      },
      "status":"DEVELOPMENT"
    },
    "schema":{
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "title": "My schema",
        "definitions": {},
        "properties": {
            "data": {
                "allOf": [
                    {
                        "type": "object",
                        "properties": {
                            "name": { "type": "string" }
                        }
                    }
                ]
            }
        }
    }
}
""";
    private Pair<String, String> generateSchemaIdAndBody(String tenant, String source, String entityType, int major, int minor, int patch) {
        String id = String.format(SCHEMA_ID, tenant, source, entityType, major, minor, patch);
        String body = String.format(SCHEMA_BODY, tenant, source, entityType, major, minor, patch, id);
        return Pair.of(id, body);
    }

    public SchemaPubSubMessage createSchemaForTest() {
        return new SchemaPubSubMessage(createSchema(), "create");
    }

    public String createSchema() {
        long millis = Instant.now().toEpochMilli();
        String entityType = String.format("test-entity-%d", millis);
        Pair<String, String> idAndBody = generateSchemaIdAndBody(TestUtils.getOsduTenant(), "test", entityType, 0, 0, 0);
        ClientResponse response = sendRequest(String.format("%s/schema", schemaApiBaseUrl), "POST", true, idAndBody.getRight(), null);

        InputStream inputStream = response.getEntityInputStream();
        String body = new BufferedReader(new InputStreamReader(inputStream, StandardCharsets.UTF_8)).lines().collect(Collectors.joining("\n"));
        logger.info("Schema service returned the following");
        logger.info(String.format("Full schema body: %s", idAndBody.getRight()));
        logger.info(body);
        assertEquals(201, response.getStatus());
        return idAndBody.getLeft();
    }

    public void validate(SchemaPubSubMessage expected, List<Message> receivedMessages) throws Exception {
        assertEquals(1, receivedMessages.size());
        logger.info(String.format("Schema message: `%s`", receivedMessages.get(0).getBody()));
        SchemaPubSubMessage actual = AwsTestUtils.unwrapSingleFirst(receivedMessages, SchemaPubSubMessage.class);
        assertEquals(expected.getKind(), actual.getKind());
        assertEquals(expected.getOp(), actual.getOp());
    }

    private ClientResponse sendRequest(String url, String method, boolean addHeaders, Object body, MediaType contentType) {
        WebResource resource = restClient.resource(url);
        WebResource.Builder resourceBuilder = resource.getRequestBuilder();
        if (contentType != null) {
            resourceBuilder.accept(contentType);
            resourceBuilder.type(contentType);
        }
        if (addHeaders) {
            dpsHeaders.getHeaders().forEach(resourceBuilder::header);
        }
        return (body == null) ? resourceBuilder.method(method, ClientResponse.class) :  resourceBuilder.method(method, ClientResponse.class, body);
    }
}
