/* Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.opengroup.osdu.notification.subscriptions;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import com.amazonaws.services.sqs.model.Message;
import org.opengroup.osdu.core.common.http.json.HttpResponseBodyMapper;
import org.opengroup.osdu.core.common.model.indexer.OperationType;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.storage.PubSubInfo;
import org.opengroup.osdu.core.common.model.storage.Record;
import org.opengroup.osdu.core.common.model.storage.StorageException;
import org.opengroup.osdu.core.common.storage.IStorageService;
import org.opengroup.osdu.core.common.storage.StorageAPIConfig;
import org.opengroup.osdu.core.common.storage.StorageFactory;
import org.opengroup.osdu.notification.util.AwsTestUtils;
import org.opengroup.osdu.notification.util.Config;

import java.util.HashSet;
import java.util.List;

public class StorageTestUtils {

    private final IStorageService storageService;
    private final HashSet<OperationType> allowedOps = new HashSet<>();
    private final String recordKind;

    public StorageTestUtils(DpsHeaders headers, HttpResponseBodyMapper mapper, SchemaTestUtils schemaTestUtils) {
        StorageAPIConfig config = StorageAPIConfig.builder().rootUrl(Config.Instance().StorageServicePath).build();
        StorageFactory subscriptionFactory = new StorageFactory(config, mapper);
        storageService = subscriptionFactory.create(headers);
        allowedOps.add(OperationType.create);
        allowedOps.add(OperationType.delete);
        recordKind = schemaTestUtils.createSchema();
    }

    private static final String RECORD_ID_PREFIX_TEST = TenantUtils.getTenantName(true) + ":query:";
    public PubSubInfo createAndPutRecord(String correlationId, String legalTag) throws StorageException {
        String recordId = RECORD_ID_PREFIX_TEST + correlationId;
        Record[] records = RecordUtils.createRecordsWithReference(1, recordId, recordKind, legalTag, true);
        storageService.upsertRecord(records);
        // We append the index in RecordUtils to ensure the IDs are unique. Maybe we don't need to do that.
        return PubSubInfo.builder().id(records[0].getId()).kind(recordKind).op(OperationType.create).build();
    }

    public void assertFirstMessagesSimilar(PubSubInfo expected, List<Message> receivedMessages) throws Exception {
        System.out.printf("Total received messages: %d\n", receivedMessages.size());
        PubSubInfo actual = AwsTestUtils.unwrapFirst(receivedMessages, PubSubInfo[].class);
        assertTrue("Expect the op to be in the \"created\" or \"deleted\" set.",  allowedOps.contains(actual.getOp()));
        assertEquals(expected.getKind(), actual.getKind());
        assertEquals(expected.getId(), actual.getId());
    }
}
