/* Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.opengroup.osdu.notification.subscriptions;

import com.amazonaws.services.simplesystemsmanagement.AWSSimpleSystemsManagement;
import com.amazonaws.services.simplesystemsmanagement.AWSSimpleSystemsManagementClientBuilder;
import com.amazonaws.services.simplesystemsmanagement.model.GetParametersRequest;
import com.amazonaws.services.simplesystemsmanagement.model.GetParametersResult;
import com.amazonaws.services.sqs.AmazonSQS;
import com.amazonaws.services.sqs.AmazonSQSClientBuilder;
import com.amazonaws.services.sqs.model.Message;
import com.amazonaws.services.sqs.model.PurgeQueueRequest;
import com.amazonaws.services.sqs.model.ReceiveMessageRequest;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.junit.After;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;
import org.opengroup.osdu.core.common.http.json.HttpResponseBodyMapper;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.notification.HmacSecret;
import org.opengroup.osdu.core.common.model.notification.Secret;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.core.common.model.storage.PubSubInfo;
import org.opengroup.osdu.core.common.notification.ISubscriptionService;
import org.opengroup.osdu.core.common.notification.SubscriptionAPIConfig;
import org.opengroup.osdu.core.common.notification.SubscriptionException;
import org.opengroup.osdu.core.common.notification.SubscriptionFactory;
import org.opengroup.osdu.notification.subscriptions.SchemaTestUtils.SchemaPubSubMessage;
import org.opengroup.osdu.notification.util.*;

import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Base64;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

public class TestNotificationsEndpoint {

    private static ISubscriptionService subscriptionService;
    private static StorageTestUtils storageUtils;
    private static LegalTagUtils legalUtils;
    private static FileTestUtils fileUtils;
    private static SchemaTestUtils schemaTestUtils;

    private static final int FILE_QUEUE_NUM = 0;
    private static final int SCHEMA_QUEUE_NUM = 0;
    private static final int INDEXER_STORAGE_INDEXER_QUEUE_NUM = 0;
    private static final int INDEXER_STORAGE_STORAGE_QUEUE_NUM = 1;
    private static final int LEGAL_STORAGE_STORAGE_QUEUE_NUM = 0;
    private static final int LEGAL_STORAGE_LEGAL_QUEUE_NUM = 1;
    private static final int REQUIRED_QUEUES = 2;

    protected static final String PARTITION_TEST = "int-test-notification";
    private static final String FILE_SERVICE_NAME = "file";
    private static final String INDEXER_SERVICE_NAME = "indexer";
    private static final String LEGAL_SERVICE_NAME = "legal";
    private static final String STORAGE_SERVICE_NAME = "storage";
    private static final String SCHEMA_SERVICE_NAME = "schema";
    private final Object mutex = new Object();
    private static final String FILE_LEGAL_TAG_NAME = String.format("%s-file-testing-%d", TenantUtils.getTenantName(true), System.currentTimeMillis());
    private static final String INDEXER_LEGAL_TAG_NAME = String.format("%s-indexer-testing-%d", TenantUtils.getTenantName(true), System.currentTimeMillis());
    private static final String STORAGE_LEGAL_TAG_NAME = String.format("%s-legal-storage-%d", TenantUtils.getTenantName(true), System.currentTimeMillis());

    private static AmazonSQS sqsClient;

    private static final Map<String, String> serviceTopicLookup = new HashMap<>();

    private final static ObjectMapper objectMapper = new ObjectMapper();

    private static final int MAX_RETRIES = 5;
    private static final int MAX_POLL_TIME = 240;


    private static class PushQueue {
        private final String queueUrl;
        private final String endpointUrl;
        private final int queueNum;
        private final Set<String> subscriptions = new HashSet<>();

        private final List<Message> messages = new ArrayList<>();

        public PushQueue(String queueUrl, String endpointUrl, int queueNum) {
            this.queueUrl = queueUrl;
            this.endpointUrl = endpointUrl;
            this.queueNum = queueNum;
        }

        public void register(String serviceName) throws SubscriptionException, InterruptedException {
            String osduTopic = serviceTopicLookup.get(serviceName);
            assertNotEquals(null, osduTopic);

            Subscription subscription = new Subscription();
            subscription.setPushEndpoint(endpointUrl);
            subscription.setSecret(subscriptionSecret);
            subscription.setId(String.format("%s-%d", osduTopic, queueNum));
            subscription.setName(String.format("%s on queue %d", osduTopic, queueNum));
            subscription.setDescription(String.format("Testing service %s osduTopic by using push endpoint to queue %d.", serviceName, queueNum));
            subscription.setTopic(osduTopic);
            String subscriptionId = null;
            int i = 0;
            boolean errored = true;
            while (errored) {
                try {
                    subscriptionId = subscriptionService.create(subscription).getId();
                    i = 0;
                    errored = false;
                } catch (SubscriptionException e) {
                    if (i >= MAX_RETRIES || e.getHttpResponse().getResponseCode() != 400)
                        throw e;
                    Thread.sleep((int) (200 * (2 << i)));
                }
                ++i;
            }
            subscriptions.add(subscriptionId);
        }

        public void pollMessages(int numMessages, int timeout) {
            messages.clear();
            ReceiveMessageRequest request = new ReceiveMessageRequest();
            request.setQueueUrl(queueUrl);
            request.setMaxNumberOfMessages(numMessages);
            final int MAX_TIMEOUT = 20;
            request.setWaitTimeSeconds(MAX_TIMEOUT);
            while (messages.size() < numMessages && timeout > MAX_TIMEOUT) {
                messages.addAll(sqsClient.receiveMessage(request).getMessages());
                timeout -= MAX_TIMEOUT;
            }
            if (messages.size() < numMessages)
                messages.addAll(sqsClient.receiveMessage(request.withWaitTimeSeconds(timeout)).getMessages());
        }

        public void pollMessages(int numMessages) {
            pollMessages(numMessages, MAX_POLL_TIME);
        }

        public List<Message> getMessages() {
            return new ArrayList<>(messages);
        }

        public void clearLocalMessages() {
            messages.forEach(message -> sqsClient.deleteMessage(queueUrl, message.getReceiptHandle()));
            messages.clear();
        }

        public void removeAllKnownSubscriptions() {
            clearLocalMessages();
            subscriptions.forEach(TestNotificationsEndpoint::ensureSubscriptionDeleted);
            subscriptions.clear();
        }

        public void purge() {
            removeAllKnownSubscriptions();
            PurgeQueueRequest purgeQueueRequest = new PurgeQueueRequest();
            purgeQueueRequest.setQueueUrl(queueUrl);
            sqsClient.purgeQueue(purgeQueueRequest);
        }

        public void removeAllPotentialSubscriptions(Map<String, String> allTopics) {
            allTopics.forEach((key, value) -> {
                String id = Base64.getEncoder().encodeToString((value + PARTITION_TEST + endpointUrl).getBytes());
                ensureSubscriptionDeleted(id);
            });
        }
    }

    private record IndexerProgress(int statusCode, String[] trace, String lastUpdateTime) {
    }

    private record IndexerMessage(String id, String kind, String operationType, String status, IndexerProgress indexProgress) {
    }


    private static String getEnv(String envVariableName) {
        String envVar = System.getenv(envVariableName);
        if (envVar == null) {
            throw new IllegalArgumentException(String.format("Must have environment variable '%s' set!", envVariableName));
        }

        return envVar;
    }

    private static List<PushQueue> queues;

    private static Secret subscriptionSecret;

    @BeforeClass
    public static void classSetup() throws Exception {
        SubscriptionAPIConfig config = SubscriptionAPIConfig.builder().rootUrl(Config.Instance().RegisterServicePath).build();
        SubscriptionFactory subscriptionFactory = new SubscriptionFactory(config);
        HttpResponseBodyMapper responseBodyMapper = new HttpResponseBodyMapper(objectMapper);


        sqsClient = AmazonSQSClientBuilder.defaultClient();

        getSubscriptionSecret();

        // Set up the subscription service used by this test
        AwsTestUtils testUtils = new AwsTestUtils();

        DpsHeaders dpsHeaders = testUtils.getDpsHeaders();
        // Set up the legal service used by this test suite
        schemaTestUtils = new SchemaTestUtils(dpsHeaders);
        storageUtils = new StorageTestUtils(dpsHeaders, responseBodyMapper, schemaTestUtils);
        legalUtils = new LegalTagUtils(dpsHeaders, responseBodyMapper);
        subscriptionService = subscriptionFactory.create(dpsHeaders);
        fileUtils = new FileTestUtils(dpsHeaders);

        getTopicNames();
        createQueues();
        purgeQueues();
        assertTrue(String.format("Must have at least %d Push Endpoint Queues to run this test!", REQUIRED_QUEUES), queues.size() >= REQUIRED_QUEUES);
    }

    private static void getSubscriptionSecret() {
        // Get the Push Endpoint Secret. Only one secret is used and shared across because otherwise, we would need to have a separate authorization for each queue, and warming up the
        // lambda functions would need to
        String pushEndpointSecret = getEnv("HMAC_SECRET");
        HmacSecret hmacSecret = new HmacSecret();
        hmacSecret.setValue(pushEndpointSecret);
        subscriptionSecret = hmacSecret;
        System.out.printf("TestNotificationsEndpoint.classSetup has HMAC Secret: %s\n", subscriptionSecret);
    }

    private static void purgeQueues() throws InterruptedException {
        queues.forEach(PushQueue::purge);
        // Wait for the queues to be purged before continuing the tests.
        Thread.sleep(60000);
    }

    private static void createQueues() {
        // Create the Push Endpoint Queues to ensure the right notifications are sent to the right queues.
        String numQueuesAsString = getEnv("PUSH_ENDPOINT_NUM_QUEUES");
        int numQueues = Integer.parseInt(numQueuesAsString);
        queues = new ArrayList<>(numQueues);
        for (int i = 0; i < numQueues; ++i) {
            String endpointUrl = getEnv(String.format("PUSH_ENDPOINT_URL_%d", i));
            String queueUrl = getEnv(String.format("PUSH_ENDPOINT_QUEUE_%d", i));
            PushQueue queue = new PushQueue(queueUrl, endpointUrl, i);
            queue.removeAllPotentialSubscriptions(serviceTopicLookup);
            queues.add(queue);
        }
    }

    private static void getTopicNames() {
        // Get the OSDU topics for each of the services that emit notifications.
        String instanceName = getEnv("OSDU_INSTANCE_NAME");
        String[] services = new String[] {FILE_SERVICE_NAME, INDEXER_SERVICE_NAME, LEGAL_SERVICE_NAME, STORAGE_SERVICE_NAME, SCHEMA_SERVICE_NAME};
        List<String> ssmNames = new ArrayList<>(services.length);
        Map<String, String> parameterReverseLookup = new HashMap<>(services.length);
        Arrays.stream(services).forEach(serviceName -> {
            String ssmParam = String.format("/osdu/instances/%s/core/%s/osdu-topic-name", instanceName, serviceName);
            ssmNames.add(ssmParam);
            parameterReverseLookup.put(ssmParam, serviceName);
        });
        AWSSimpleSystemsManagement ssm = AWSSimpleSystemsManagementClientBuilder.defaultClient();
        GetParametersRequest getParametersRequest = new GetParametersRequest().withNames(ssmNames);
        GetParametersResult getParametersResult = ssm.getParameters(getParametersRequest);
        getParametersResult.getParameters().forEach(parameter -> {
            String serviceName = parameterReverseLookup.get(parameter.getName());
            serviceTopicLookup.put(serviceName, parameter.getValue());
        });
    }

    private static void ensureSubscriptionDeleted(String subscriptionId) {
        try {
            subscriptionService.delete(subscriptionId);
        }
        catch (Exception e) {
            // Don't care, because the subscription may or may not exist.
            // If it doesn't exist, then this will erroneously error out if we don't catch it first.
        }
    }

    @After
    public void tearDown() throws Exception {
        queues.forEach(PushQueue::removeAllKnownSubscriptions);
        legalUtils.ensureDeleted(STORAGE_LEGAL_TAG_NAME);
        legalUtils.ensureDeleted(INDEXER_LEGAL_TAG_NAME);
        legalUtils.ensureDeleted(FILE_LEGAL_TAG_NAME);
        Thread.sleep(10000); // Give time for the legal notifications to be sent...
        queues.forEach(queue -> queue.removeAllPotentialSubscriptions(serviceTopicLookup));
        purgeQueues();
    }

    private List<List<Message>> pollForMessages(int timeout, int numMessages, PushQueue... queues) throws InterruptedException {
        List<Thread> threads = new ArrayList<>(queues.length);
        for (PushQueue queue : queues) {
            Runnable pollMessagesRunnable = () -> queue.pollMessages(numMessages);
            if (timeout >= 0) {
                pollMessagesRunnable = () -> queue.pollMessages(numMessages, timeout);
            }
            Thread thread = new Thread(pollMessagesRunnable);
            thread.start();
            threads.add(thread);
        }
        for (Thread thread : threads) {
            thread.join();
        }
        List<List<Message>> messages = new ArrayList<>(queues.length);
        for (PushQueue queue : queues) {
            messages.add(queue.getMessages());
        }
        return messages;
    }

    private List<List<Message>> pollForMessages(PushQueue... queues) throws InterruptedException {
        return pollForMessages(-1, 1, queues);
    }

    @Test
    public void test_File_Notifications_Received() throws Exception {
        synchronized (mutex) {
            PushQueue fileQueue = queues.get(FILE_QUEUE_NUM);
            try {
                fileQueue.register(FILE_SERVICE_NAME);
                legalUtils.addLegalTag(FILE_LEGAL_TAG_NAME);
                Instant start = Instant.now();
                String fileId = fileUtils.uploadFile(FILE_LEGAL_TAG_NAME);

                // Get Queue Messages
                List<List<Message>> messagesLists = pollForMessages(-1, 3, fileQueue);
                assertEquals(1, messagesLists.size());
                List<Message> fileSQSMessages = messagesLists.get(0);

                fileUtils.compareMessagesWithExpected(fileId, fileSQSMessages, start);
            } catch (Exception e) {
                System.err.println(e.toString());
                e.printStackTrace();
                throw e;
            } finally {
                legalUtils.ensureDeleted(FILE_LEGAL_TAG_NAME);
            }
        }
    }

    @Test
    public void test_Schema_Notifications_Received() throws Exception {
        synchronized (mutex) {
            PushQueue schemaQueue = queues.get(SCHEMA_QUEUE_NUM);
            try {
                schemaQueue.register(SCHEMA_SERVICE_NAME);
                SchemaPubSubMessage expected = schemaTestUtils.createSchemaForTest();
                List<List<Message>> messagesLists = pollForMessages(schemaQueue);
                assertEquals(1, messagesLists.size());
                schemaTestUtils.validate(expected, messagesLists.get(0));
            } catch (Exception e) {
                System.err.println(e.toString());
                e.printStackTrace();
                throw e;
            }
        }
    }

    @Test
    public void test_Indexer_Notifications_Received() throws Exception {
        synchronized (mutex) {
            PushQueue storageQueue = queues.get(INDEXER_STORAGE_STORAGE_QUEUE_NUM);
            PushQueue indexerQueue = queues.get(INDEXER_STORAGE_INDEXER_QUEUE_NUM);
            try {
                indexerQueue.register(INDEXER_SERVICE_NAME);
                storageQueue.register(STORAGE_SERVICE_NAME);
                Instant start = Instant.now();

                legalUtils.addLegalTag(INDEXER_LEGAL_TAG_NAME);
                PubSubInfo expectedStorageMessage = storageUtils.createAndPutRecord(UUID.randomUUID().toString(), INDEXER_LEGAL_TAG_NAME);

                // Get Queue Messages
                List<List<Message>> messagesLists = pollForMessages(600, 1, storageQueue, indexerQueue);
                assertEquals(2, messagesLists.size());
                List<Message> storageSQSMessages = messagesLists.get(0);
                List<Message> indexerSQSMessages = messagesLists.get(1);

                Instant end = Instant.now();

                storageUtils.assertFirstMessagesSimilar(expectedStorageMessage, storageSQSMessages);
                IndexerMessage indexerMessage = AwsTestUtils.unwrapFirst(indexerSQSMessages, IndexerMessage[].class);
                assertEquals(expectedStorageMessage.getKind(), indexerMessage.kind());
                assertEquals(expectedStorageMessage.getId(), indexerMessage.id());
                assertEquals("create", indexerMessage.operationType());
                assertEquals("SUCCESS", indexerMessage.status());
                assertNotNull(indexerMessage.indexProgress());
                assertEquals(200, indexerMessage.indexProgress().statusCode());
                Instant lastChanged = Instant.parse(indexerMessage.indexProgress().lastUpdateTime());
                assertTrue(lastChanged.isAfter(start) && lastChanged.isBefore(end));
            } catch (Exception e) {
                System.err.println(e.toString());
                e.printStackTrace();
                throw e;
            } finally {
                legalUtils.ensureDeleted(INDEXER_LEGAL_TAG_NAME);
            }
        }
    }

    @Test
    public void test_LegalAndStorage_Notifications_Received() throws Exception {
        synchronized (mutex) {
            PushQueue legalQueue = queues.get(LEGAL_STORAGE_LEGAL_QUEUE_NUM);
            PushQueue storageQueue = queues.get(LEGAL_STORAGE_STORAGE_QUEUE_NUM);
            try {
                legalQueue.register(LEGAL_SERVICE_NAME);
                storageQueue.register(STORAGE_SERVICE_NAME);

                legalUtils.addLegalTag(STORAGE_LEGAL_TAG_NAME);

                final String correlationId = UUID.randomUUID().toString();
                PubSubInfo expectedStorageMessage = storageUtils.createAndPutRecord(correlationId, STORAGE_LEGAL_TAG_NAME);
                List<List<Message>> messagesLists = pollForMessages(storageQueue);
                assertEquals(1, messagesLists.size());
                List<Message> storageSQSMessages = messagesLists.get(0);
                storageUtils.assertFirstMessagesSimilar(expectedStorageMessage, storageSQSMessages);
                storageQueue.clearLocalMessages();

                LegalMessage expectedLegalMessage = legalUtils.deleteLegalTag(STORAGE_LEGAL_TAG_NAME);

                // Get Queue Messages
                messagesLists = pollForMessages(storageQueue, legalQueue);
                assertEquals(2, messagesLists.size());
                storageSQSMessages = messagesLists.get(0);
                List<Message> legalSQSMessages = messagesLists.get(1);

                storageUtils.assertFirstMessagesSimilar(expectedStorageMessage, storageSQSMessages);
                legalUtils.assertFirstMessagesSimilar(expectedLegalMessage, legalSQSMessages);
            } catch (Exception e) {
                System.err.println(e.toString());
                e.printStackTrace();
                throw e;
            } finally {
                legalUtils.ensureDeleted(STORAGE_LEGAL_TAG_NAME);
            }
        }
    }
}
